from flux.console.screens.dashboard import DashboardView
from flux.console.screens.workflows import WorkflowsView
from flux.console.screens.executions import ExecutionsView
from flux.console.screens.workers import WorkersView
from flux.console.screens.schedules import SchedulesView
from flux.console.screens.logs import LogsView

__all__ = [
    "DashboardView",
    "WorkflowsView",
    "ExecutionsView",
    "WorkersView",
    "SchedulesView",
    "LogsView",
]
